/*
*  ALCATEL SpeedTouch USB modem microcode extract utility
*  Copyright (C) 2001 Benoit PAPILLAULT
*  
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*  Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
*  Creation : 25/04/2001
*
*  Searching for the microcode is done using the best match for a start and end 
*  patterns. This has been tested with windows 1.3 drivers, linux 1.3 drivers
*  & linux 1.3.2 drivers (using binary available at the Alcatel website:
*  http://www.alcateldsl.com/ ).
*
* $Id$
*/

#ifndef _EXTRACT_C_
#define _EXTRACT_C_

#include "extract.h"
#include "modem.h"
#include "pppoa3.h"
#include "crc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#undef STANDALONE_EXTRACTER

/*****************************************************************************
* Local data
*****************************************************************************/

typedef struct _firmware_id_t
{
	unsigned long crc;
	unsigned long size;
	const char *id;
}firmware_id_t;

static const firmware_id_t firmware_ids[] =
{
	{0xae3ff81f, 526239, "1.3.1 - GNU/Linux"},
	{0xa719bc0e, 523436, "1.3.1 - Win32"},
	{0x94a45435, 526187, "1.3.3 - GNU/Linux - Win32"},
	{0x61914198, 527093, "1.4.0 - Win32(3.0.1800)"},
	{0x37c189ed, 527752, "1.4.0 - Win32(4.0.100)"},
	{0x99cc1c1a, 528738, "1.4.2 - Win32(2.0.500)"},
	{0xe0251a5e, 671665, "1.6.1 - Win32(5.0.1801)"},
	{0x3b4a5854, 671653, "1.6.1 - MacOSX - Win32(1.0.1800)"},
	{0xd673923f, 672264, "2.0.0 - Win32(07)"},
	{0x5bca7d16, 677641, "2.0.1 - MacOSX - Win32(2.0.0)"},
};

/*****************************************************************************
* Local Prototypes
*****************************************************************************/

static unsigned char *load_file(const char * file, int * size);

static int search_match(
	unsigned char *buffer,
	int buffer_length,
	unsigned char *patterns,
	int pattern_length,
	int nb_patterns);

/******************************************************************************
*	Main Lib Function
******************************************************************************/

/*
* Function     : extract_microcode
* Return Value : NULL on error
*                valid pointer to the microcode on success
* Description  : We are searching for the best match of a start and end
*                sequence.
*                Those sequence delimits the microcode. You won't get 100%
*                probability, but it will work anyway, for all the microcode
*                I've tested.
*/
unsigned char *extract_microcode(const char *file, int * size)
{
	/* Variables */
	int start;
	unsigned char * buf1;
	int size1;

	/* Start and end sequences of bytes of firmware */
#define PATTERN_LENGTH 8
	unsigned char start_pattern[PATTERN_LENGTH] = {
		0x88, 0xf8, 0x00, 0x40, 0x00, 0x00, 0xf2, 0x01
	};

	unsigned char stop_pattern[PATTERN_LENGTH] = {
		0x04, 0x00, 0x40, 0x00, 0x00, 0x00, 0x01, 0x14
	};

	/* Loads the firmware file */
	if((buf1 = load_file(file,&size1)) == NULL)
		return(NULL);

	/* Searches the start and end offsets */
	start = search_match(buf1, size1, start_pattern, PATTERN_LENGTH, sizeof(start_pattern)/PATTERN_LENGTH);

	if(start < 0) {
		free(buf1);
		return(NULL);
	}

	*size  = search_match(buf1 + start, size1 - start, stop_pattern , PATTERN_LENGTH, sizeof(stop_pattern)/PATTERN_LENGTH);

	if (*size < 0) {
		free(buf1);
		return(NULL);
	}

	/* We must add the pattern length to obtain the real length */
	*size += PATTERN_LENGTH;

	/* Move the data to begining of the buffer */
	memcpy(buf1, buf1 + start, *size);
	memset(buf1 + *size, 0, size1 - *size);

#ifndef STANDALONE_EXTRACTER
	{
		int loop;
		const char *id = "Unknown revision - Please report the CRC and size with the revision number to speedtouch@ml.free.fr";

		for(loop=0; loop<sizeof(firmware_ids)/sizeof(firmware_ids[0]); loop++) {
			if(firmware_ids[loop].crc == ~aal5_calc_crc(buf1, *size,~0) &&
			   firmware_ids[loop].size == *size)
				id = firmware_ids[loop].id;
		}

		report(1, REPORT_INFO, "Firmware information (CRC:0x%8x, Size:%d, %s)\n",
		       ~aal5_calc_crc(buf1, *size,~0),
		       *size,
		       id);
	}


#endif

	/* Returns a pointer to the start adress */
	return(buf1);

}

/*****************************************************************************
*	Local sub routines
*****************************************************************************/

/*
* Function     : load_file
* Return Value : null in case of error
*                a pointer to a buffer allocated using malloc() on success
* NB           : *size contains the returned buffer size
*/

static unsigned char * load_file(const char * file, int * size)
{
	struct stat statbuf;
	int len ;
	unsigned char * buf;
	int fd;

	/* Opens the file */
	if ((fd = open(file,O_RDONLY)) < 0) {
		perror(file);
		return(NULL);
	}

	/* Retrieves file informations */
	 if (fstat(fd,&statbuf) != 0) {
		perror("stat");
		return(NULL);
	}

	/* Gets length */
	len = statbuf.st_size;

	/* Allocates the buffer */
	if((buf = (unsigned char *)malloc(len)) == NULL) {
		perror("malloc");
		close (fd);
		return(NULL);
	}

	/* Read all contents in buffer */
	if (read(fd,buf,len) != len) {
		perror(file);
		free (buf);
		close (fd);
		return(NULL);
	}

	/* Closes the file */
	close(fd);

	/* Sets size */
	*size = len;

	return(buf);

}

/*
* Function     : search_match
* Return value : -1 in case of error
*                best match offset on success
* Description  : Search for the longest match of buf2 inside buf1.
*
*/
static int search_match(
	unsigned char *buffer,
	int buffer_length,
	unsigned char *patterns,
	int pattern_length,
	int nb_patterns)
{

	int cur_offset;
	int pattern;

	/* Number of bytes match with pattern */
	int best_match = 0;

	/* Best offset*/
	int best_offset = -1;

	for(pattern=0; pattern<nb_patterns; pattern++) {

		for(cur_offset=0;cur_offset< (buffer_length - pattern_length + 1);cur_offset++) {

			int i;
			int cur_match;

			/* Computing cur_match */
			cur_match = 0;

			for(i=0;i<pattern_length;i++)
				if (buffer[cur_offset+i] == patterns[i])
					cur_match ++;

			if(cur_match > best_match) {

				best_offset = cur_offset;
				best_match  = cur_match;

				if (best_match == pattern_length)
					goto best_offset_found;

			}

		}

		patterns += pattern_length;

	}

 best_offset_found:
		
#ifndef STANDALONE_EXTRACTER
	report(1, REPORT_INFO, "Best offset %6d with probability %3d%%\n",
	       best_offset,
	       (100*best_match)/pattern_length);
#endif

	return(best_offset);

}

/******************************************************************************
*	Standalone part of the extract utility
******************************************************************************/

/*
* NB : As we don't have right to provide any microcode
*      (Read the Alcatel's License for details)
*      we decided to link our modem_run to it at runtime
*      so we don't dissamble|modify|rev engineer the firmware
*      It can be considered as a lib, or as a plugin for our
*      driver... Our dream is to provide you a totally free
*      driver with a complete GPL firmware too :-)
*/

#ifdef STANDALONE_EXTRACTER

#include "crc.c"

void usage(char *progname)
{
	fprintf(stderr,	"Usage: %s <source> <dest>\n", progname);
	exit (-1);
}

int main(int argc, char *argv[])
{
	const char * source = argv[1];
	const char * dest   = argv[2];

	int size;
	int fd;
	unsigned char * buf;

	if (argc != 3)
		usage(argv[0]);

	buf = extract_microcode(source, &size);
	if (buf == NULL) {
		fprintf(stderr, "Error: microcode not found!\n");
		return(-1);
	}
	
	if((fd = open(dest ,O_RDWR|O_CREAT|O_TRUNC,0644)) < 0) {
		perror("open");
		return(-1);
	}
	
	if (write(fd, buf, size) != size) {
		perror("write");
		return(-1);
	}
	
	printf("Firmware CRC: 0x%8x\n", ~aal5_calc_crc(buf, size,~0));
	printf("Firmware Size: %d\n", size);
		
	free(buf);
	
	close(fd);

	return(0);

}

#endif /* STANDALONE_EXTRACTER */

#endif /* #ifndef _EXTRACT_C_ */
